#include "protocol.h"

#include <stdlib.h>
#include <string.h>
#include <sys/time.h>

#include "player.h"
#include "server.h"
#include "udp.h"

// TODO: test this function
int difference_ignoring_wraparound(int a, int b, int max_length) {
  int diff = b - a;
  if (diff > max_length / 2 || diff < -max_length / 2) {
    a = (a + (max_length / 2)) % max_length;
    b = (b + (max_length / 2)) % max_length;
    return b - a;
  } else {
    return b - a;
  }
}

int handle_packet(Player* player, char* data, int length) {
  int pos = 0;

  unsigned short packet_length;
  READ_SHORT(packet_length);
  if (packet_length > length) return packet_length;

  unsigned short sequence;
  READ_SHORT(sequence);
  int diff = difference_ignoring_wraparound(player->ack, sequence, USHRT_MAX);
  if (diff >= 0) {
    player->ack_field <<= diff;
    player->ack_field |= 1;
    player->ack = sequence;
  } else if (-diff < 32) {
    player->ack_field |= (1 << -diff);
  }

  // Ack unacked packets that match the given ack field
  unsigned short ack;
  READ_SHORT(ack);
  unsigned int ack_field;
  READ_INT(ack_field);
  for (int i = 0; i < 32; i++) {
    if ((ack_field >> i) & 1) {
      int acked_sequence = (ack - i) % USHRT_MAX;
      for (int j = 0; j < player->nb_unacked_packets; j++) {
        if (player->unacked_packets[j].sequence == acked_sequence) {
          uv_timer_stop(&player->unacked_packets[j].timer);
          free(player->unacked_packets[j].buffer.base);

          struct timespec send_time = player->unacked_packets[j].send_time;
          struct timespec recv_time;
          clock_gettime(CLOCK_MONOTONIC_COARSE, &recv_time);
          uint64_t rtt = (recv_time.tv_sec - send_time.tv_sec) * 1000000 +
                         (recv_time.tv_nsec - send_time.tv_nsec) / 1000;
          player->rtt += (rtt - player->rtt) / 10;

          if (j + 1 < player->nb_unacked_packets) {
            memmove(&player->unacked_packets[j],
                    &player->unacked_packets[j + 1], sizeof(UnackedPacket));
          }
          player->nb_unacked_packets--;
          break;
        }
      }
    }
  }

  unsigned short packet_type;
  READ_SHORT(packet_type);

  if (player->username[0] == '\0' && packet_type != C_PACKET_LOGIN) {
    // Ignore packet
    return packet_length + sizeof(unsigned short);
  }

  // TODO: reading directly into the player struct might cause some unwanted
  // behavior if the packets aren't complete. read into temporary variables
  // if needed.

  switch (packet_type) {
    case C_PACKET_LOGIN: {
      READ_STRING(player->username);

      // Pick unused player id
      static unsigned short new_player_id = 0;
      new_player_id = new_player_id + 1 % 32767;
      int new_id_found = 0;
      while (!new_id_found) {
        new_id_found = 1;
        for (int i = 0; i < MAX_PLAYERS; i++) {
          if (players[i].id == new_player_id) {
            // '32767' is max unsigned short value
            new_player_id = new_player_id + 1 % 32767;
            new_id_found = 0;
            break;
          }
        }
      }

      player->id = new_player_id;

      INIT_PACKET(S_PACKET_LOGIN);
      WRITE_BYTE(0);
      IMPORTANT_PACKET_SEND(player);

      return pos;
    }

    case C_PACKET_HELLO: {
      READ_STRING(player->username);
      READ_STRING(player->scene);
      READ_FLOAT(player->x);
      READ_FLOAT(player->y);
      READ_BOOL(player->scale);
      READ_SHORT(player->animation);

      INIT_PACKET(S_PACKET_UPDATE_SETTINGS);
      WRITE_BOOL(PVP_ENABLED);
      WRITE_BOOL(BODY_DAMAGE_ENABLED);
      WRITE_BOOL(ALWAYS_SHOW_MAP_ICONS);
      WRITE_BOOL(ONLY_SHOW_MAP_ICONS_WITH_WAYWARD_COMPASS);
      WRITE_BOOL(DISPLAY_NAMES);
      WRITE_BOOL(TEAMS_ENABLED);
      WRITE_BOOL(ALLOW_SKINS);
      WRITE_BYTE(NAIL_DAMAGE);
      WRITE_BYTE(GRUBBERFLY_ELEGY_DAMAGE);
      WRITE_BYTE(VENGEFUL_SPIRIT_DAMAGE);
      WRITE_BYTE(SHADE_SOUL_DAMAGE);
      WRITE_BYTE(DESOLATE_DIVE_DAMAGE);
      WRITE_BYTE(HOWLING_WRAITH_DAMAGE);
      WRITE_BYTE(ABYSS_SHRIEK_DAMAGE);
      WRITE_BYTE(GREAT_SLASH_DAMAGE);
      WRITE_BYTE(DASH_SLASH_DAMAGE);
      WRITE_BYTE(CYCLONE_SLASH_DAMAGE);
      WRITE_BYTE(SPORE_SHROOM_DAMAGE);
      WRITE_BYTE(SPORE_DUNG_SHROOM_DAMAGE);
      WRITE_BYTE(THORN_OF_AGONY_DAMAGE);
      IMPORTANT_PACKET_SEND(player);

      {
        INIT_PACKET(S_PACKET_CONNECT);
        WRITE_SHORT(player->id);
        WRITE_STRING(player->username);
        IMPORTANT_SERVER_BROADCAST(player);
      }

      add_player_to_scene(player);

      return pos;
    }

    case C_PACKET_DISCONNECT: {
      kick_player(player, 0);
      return pos;
    }

    case C_PACKET_UPDATE_PLAYER: {
      unsigned char nb_animations = 0;

      // This will get a bit confusing, but we're writing the packet we're
      // going to broadcast while still reading the received packet.
      INIT_PACKET(S_PACKET_UPDATE_PLAYER);

      unsigned short unused_player_id;
      READ_SHORT(unused_player_id);
      WRITE_SHORT(player->id);

      unsigned char update_flag;
      READ_BYTE(update_flag);
      WRITE_BYTE(update_flag);

      if (update_flag & (1 << PLAYER_UPDATE_POSITION)) {
        READ_FLOAT(player->x);
        READ_FLOAT(player->y);
        WRITE_FLOAT(player->x);
        WRITE_FLOAT(player->y);
      }

      if (update_flag & (1 << PLAYER_UPDATE_SCALE)) {
        READ_BOOL(player->scale);
        WRITE_BOOL(player->scale);
      }

      if (update_flag & (1 << PLAYER_UPDATE_MAP_POSITION)) {
        READ_FLOAT(player->map_x);
        READ_FLOAT(player->map_y);
        WRITE_FLOAT(player->map_x);
        WRITE_FLOAT(player->map_y);

        // TODO: if(ALWAYS_SHOW_ICONS || ONLY_SHOW_MAP_ICONS_W/_WWD_COMPASS)
        //       broadcast map position update to whole server insteaf of
        //       current scene
      }

      if (update_flag & (1 << PLAYER_UPDATE_ANIMATION)) {
        READ_BYTE(nb_animations);
        WRITE_BYTE(nb_animations);

        for (int i = 0; i < nb_animations; i++) {
          READ_SHORT(player->animation);
          WRITE_SHORT(player->animation);

          unsigned char frame;
          READ_BYTE(frame);
          WRITE_BYTE(frame);

          unsigned char num_effects;
          READ_BYTE(num_effects);
          WRITE_BYTE(num_effects);

          // There is one effect per bit. So read (num_effects / 8) bytes.
          for (int j = 0; j < num_effects; j += 8) {
            unsigned char tmp;
            READ_BYTE(tmp);
            WRITE_BYTE(tmp);
          }
        }
      }

      SCENE_BROADCAST(player);
    }

    case C_PACKET_UPDATE_ENTITY: {
      // TODO: confirm sender is scene host before relaying

      // This will get a bit confusing, but we're writing the packet we're
      // going to broadcast while still reading the received packet.
      INIT_PACKET(S_PACKET_UPDATE_ENTITY);

      unsigned char entity_type;
      READ_BYTE(entity_type);
      WRITE_BYTE(entity_type);

      unsigned char entity_id;
      READ_BYTE(entity_id);
      WRITE_BYTE(entity_id);

      unsigned char update_flag;
      READ_BYTE(update_flag);
      WRITE_BYTE(update_flag);

      if (update_flag & (1 << ENTITY_UPDATE_POSITION)) {
        float x, y;
        READ_FLOAT(x);
        READ_FLOAT(y);
        WRITE_FLOAT(x);
        WRITE_FLOAT(y);
      }

      if (update_flag & (1 << ENTITY_UPDATE_STATE)) {
        unsigned char state;
        READ_BYTE(state);
        WRITE_BYTE(state);
      }

      if (update_flag & (1 << ENTITY_UPDATE_VARIABLES)) {
        unsigned char num_bytes;
        READ_BYTE(num_bytes);
        WRITE_BYTE(num_bytes);

        for (int i = 0; i < num_bytes; i++) {
          unsigned char tmp;
          READ_BYTE(tmp);
          WRITE_BYTE(tmp);
        }
      }

      SCENE_BROADCAST(player);
    }

    case C_PACKET_ENTER_SCENE: {
      READ_STRING(player->scene);
      READ_FLOAT(player->x);
      READ_FLOAT(player->y);
      READ_BOOL(player->scale);
      READ_SHORT(player->animation);

      add_player_to_scene(player);

      return pos;
    }

    case C_PACKET_LEAVE_SCENE: {
      remove_player_from_scene(player);

      return pos;
    }

    case C_PACKET_DIE: {
      INIT_PACKET(S_PACKET_DIE);
      WRITE_SHORT(player->id);
      SCENE_BROADCAST(player);

      return pos;
    }

    case C_PACKET_UPDATE_TEAM: {
      READ_BYTE(player->team);

      INIT_PACKET(S_PACKET_UPDATE_TEAM);
      WRITE_SHORT(player->id);
      WRITE_STRING(player->username);
      WRITE_BYTE(player->team);
      IMPORTANT_SERVER_BROADCAST(player);

      return pos;
    }

    case C_PACKET_UPDATE_SKIN: {
      READ_BYTE(player->skin_id);

      INIT_PACKET(S_PACKET_UPDATE_SKIN);
      WRITE_SHORT(player->id);
      WRITE_BYTE(player->skin_id);
      IMPORTANT_SCENE_BROADCAST(player);

      return pos;
    }

    default:
      printf("Unhandled UDP packet type %d\n", packet_type);
      return packet_length + sizeof(unsigned short);
  }
}
