#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <uv.h>

#include "player.h"
#include "protocol.h"
#include "server.h"

uv_udp_t* hollowknight_server = NULL;

// TODO: make sure we handle libuv errors returned by all functions
// TODO: add a lot of printfs

// Schedules a packet to be sent to the player on the next server tick
void send_packet_to_player(Player* player, char* data, int len,
                           bool is_important) {
  if (is_important) {
    if (len > MAX_BYTES_PER_TICK) {
      printf(
          "Important packet larger than %d bytes (%d bytes). "
          "Player: %s\n",
          MAX_BYTES_PER_TICK, len, player->username);
    }

    player->nb_must_send_packets += 1;
    player->must_send_packets =
        reallocarray(player->must_send_packets, player->nb_must_send_packets,
                     sizeof(uv_buf_t));
    player->must_send_packets[player->nb_must_send_packets - 1] =
        uv_buf_init(data, len);
    player->outgoing_bytes += len;
  } else {
    if (player->outgoing_bytes > MAX_BYTES_PER_TICK) {
      // Data limit reached for this tick, drop the packet
      free(data);
    } else {
      player->nb_may_send_packets += 1;
      player->may_send_packets =
          reallocarray(player->may_send_packets, player->nb_may_send_packets,
                       sizeof(uv_buf_t));
      player->may_send_packets[player->nb_may_send_packets - 1] =
          uv_buf_init(data, len);
      player->outgoing_bytes += len;
    }
  }
}

void broadcast_packet(Player* source_player, char* packet, int len,
                      bool same_scene_only, bool is_important) {
  for (int i = 0; i < MAX_PLAYERS; i++) {
    if (players[i].username[0] == '\0') continue;
    if (players[i].id == source_player->id) continue;
    if (same_scene_only &&
        (strcmp(players[i].scene, source_player->scene) != 0))
      continue;

    send_packet_to_player(&players[i], packet, len, is_important);
  }
}

struct udp_send_data {
  unsigned short player_id;
  uv_udp_send_t write;
  uv_buf_t* buffers;
  int nb_buffers;
};

void end_udp_write(uv_udp_send_t* write, int status) {
  (void)status;
  struct udp_send_data* data = write->data;
  for (int i = 0; i < data->nb_buffers; i++) {
    free(data->buffers[i].base);
  }
  free(data->buffers);
  free(data);
}

void resend_lost_packet(uv_timer_t* handle) {
  UnackedPacket* up = (UnackedPacket*)handle->data;
  up->resends++;
  if (up->resends > MAX_RESENDS) {
    printf("debug: Dropping lost packet for %s\n", up->player->username);
    return;
  }

  printf("debug: Resending lost packet for %s\n", up->player->username);

  struct udp_send_data* out = malloc(sizeof(struct udp_send_data));
  out->buffers = malloc(sizeof(uv_buf_t));
  out->nb_buffers = 1;
  out->write.data = out;

  clock_gettime(CLOCK_MONOTONIC_COARSE, &up->send_time);
  uv_udp_send(&out->write, hollowknight_server, out->buffers, out->nb_buffers,
              (const struct sockaddr*)&up->player->addr, end_udp_write);
}

void process_tick_for_player(Player* p) {
  uv_loop_t* loop = uv_default_loop();

  int nb_packets = p->nb_must_send_packets + p->nb_may_send_packets;
  if (nb_packets == 0) return;

  int outgoing_bytes = 0;
  struct udp_send_data* out = malloc(sizeof(struct udp_send_data));
  out->buffers = calloc(nb_packets, sizeof(uv_buf_t));
  out->write.data = out;

  for (int j = 0; j < p->nb_must_send_packets; j++) {
    unsigned short* packet_ptr = (unsigned short*)p->must_send_packets[j].base;
    unsigned short packet_len = packet_ptr[0];

    out->buffers[j] = p->must_send_packets[j];
    out->nb_buffers++;

    packet_ptr[1] = p->local_sequence_number;
    p->local_sequence_number = (p->local_sequence_number + 1) % USHRT_MAX;

    packet_ptr[2] = p->ack;
    unsigned int* ack_field = (unsigned int*)&packet_ptr[3];
    *ack_field = p->ack_field;

    outgoing_bytes += packet_len;

    // Add copy of the packet to unacked packet list
    // It will be resent after player->rtt * 2 milliseconds unless acked
    p->unacked_packets = reallocarray(
        p->unacked_packets, p->nb_unacked_packets + 1, sizeof(UnackedPacket));
    UnackedPacket* up = &p->unacked_packets[p->nb_unacked_packets];
    p->nb_unacked_packets++;
    char* packet_cpy = malloc(p->must_send_packets[j].len);
    memcpy(packet_cpy, p->must_send_packets[j].base,
           p->must_send_packets[j].len);
    clock_gettime(CLOCK_MONOTONIC_COARSE, &up->send_time);
    up->player = p;
    up->sequence = packet_ptr[1];
    up->buffer = uv_buf_init(packet_cpy, p->must_send_packets[j].len);
    uv_timer_init(loop, &up->timer);
    up->timer.data = up;
    uv_timer_start(&up->timer, resend_lost_packet, p->rtt * 2, p->rtt * 2);

    if (p->nb_unacked_packets > MAX_UNACKED_PACKETS) {
      kick_player(p, true);
      return;
    }
  }

  for (int j = 0; j < p->nb_may_send_packets; j++) {
    unsigned short* packet_ptr = (unsigned short*)p->may_send_packets[j].base;
    unsigned short packet_len = packet_ptr[0];
    if (outgoing_bytes + packet_len > MAX_BYTES_PER_TICK) break;

    out->buffers[p->nb_must_send_packets + j] = p->may_send_packets[j];
    out->nb_buffers++;

    packet_ptr[1] = p->local_sequence_number;
    p->local_sequence_number = (p->local_sequence_number + 1) % USHRT_MAX;

    packet_ptr[2] = p->ack;
    unsigned int* ack_field = (unsigned int*)&packet_ptr[3];
    *ack_field = p->ack_field;

    outgoing_bytes += packet_len;
  }

  free(p->must_send_packets);
  p->must_send_packets = NULL;
  p->nb_must_send_packets = 0;
  free(p->may_send_packets);
  p->may_send_packets = NULL;
  p->nb_may_send_packets = 0;

  uv_udp_send(&out->write, hollowknight_server, out->buffers, out->nb_buffers,
              (const struct sockaddr*)&p->addr, end_udp_write);
}

void process_tick(uv_timer_t* handle) {
  (void)handle;

  for (int i = 0; i < MAX_PLAYERS; i++) {
    if (players[i].username[0] == '\0') continue;
    process_tick_for_player(&players[i]);
  }
}

void udp_read(uv_udp_t* req, ssize_t nread, const uv_buf_t* buf,
              const struct sockaddr* addr, unsigned flags) {
  // We need a remote IP address to do anything with the packet.
  if (addr == NULL || nread <= 0) {
    if ((flags & UV_UDP_MMSG_FREE) && buf->base) {
      free(buf->base);
    }
    return;
  }

  // Get player from UDP handle address
  req->data = NULL;
  for (int i = 0; i < MAX_PLAYERS; i++) {
    // TODO: double check ss_family vs sa_family
    if (players[i].addr.ss_family != addr->sa_family) continue;

    if (addr->sa_family == AF_INET) {
      struct sockaddr_in* one = (struct sockaddr_in*)addr;
      struct sockaddr_in* two = (struct sockaddr_in*)&players[i].addr;
      if (one->sin_addr.s_addr != two->sin_addr.s_addr) continue;
      if (one->sin_port != two->sin_port) continue;

      req->data = &players[i];
      break;
    } else if (addr->sa_family == AF_INET6) {
      struct sockaddr_in6* one = (struct sockaddr_in6*)addr;
      struct sockaddr_in6* two = (struct sockaddr_in6*)&players[i].addr;
      if (memcmp(one->sin6_addr.s6_addr, two->sin6_addr.s6_addr,
                 sizeof(one->sin6_addr.s6_addr)) != 0)
        continue;
      if (one->sin6_port != two->sin6_port) continue;

      req->data = &players[i];
      break;
    }
  }

  // No player found, pick the first available spot
  if (req->data == NULL) {
    for (int i = 0; i < MAX_PLAYERS; i++) {
      if (players[i].username[0] == '\0') {
        players[i].rtt = 200;
        req->data = &players[i];
        break;
      }
    }
  }

  if (req->data != NULL) {
    Player* player = (Player*)req->data;
    player->last_packet_tms = time(NULL);

    int pos = 0;
    while (nread - pos > 0) {
      pos += handle_packet(player, buf->base + pos, nread - pos);
    }
  }
}
